from abc import abstractmethod


class NehushtanMessageQueue:
    """
    This is a simplest model of Message Queue.
    Since 0.4.12
    Since 0.4.16 use `enqueue and dequeue` instead of `push and pop`
    """

    def __init__(self):
        pass

    @abstractmethod
    def enqueue(self, item: str, queue_name: str):
        pass

    @abstractmethod
    def dequeue(self, queue_name: str):
        pass



from abc import abstractmethod

from nehushtan.httpd.NehushtanHTTPRequestHandler import NehushtanHTTPRequestHandler


class NehushtanHTTPRequestFilter:
    """
    Since 0.4.0
    """

    def __init__(self, http_handler: NehushtanHTTPRequestHandler, extra_data_dict: dict = None):
        self.__http_handler = http_handler

        # When the request should be deny
        self.http_code_for_denial: int = 200
        self.message_for_denial: str = ''

        # When the request should be accept, something computed would be given out in the dict form
        self.extra_data_dict = extra_data_dict if extra_data_dict is not None else {}

    def get_http_handler(self):
        return self.__http_handler

    @abstractmethod
    def should_accept_request(self) -> bool:
        pass



import random
import time
from threading import Thread

from nehushtan.socket.NehushtanTCPSocketClient import NehushtanTCPSocketClient


class QAClient(NehushtanTCPSocketClient):
    def handle_client_conneciton(self):
        for i in range(10):
            r = random.Random()
            count = r.randint(2, 5)
            x = []
            for i in range(count):
                x.append(f'{r.randint(10, 20)}')
            x = '+'.join(x)

            print(f'Q: {x}=?')

            self.get_connection().send(x.encode())
            data = self.get_connection().recv(1024)

            print(f'A: {data.decode()}')

            time.sleep(1)

        self.close_connection()


def client_call():
    QAClient(4354).listen()


if __name__ == '__main__':

    for i in range(20):
        Thread(target=client_call).start()
        time.sleep(random.Random().random() * 5)



from abc import abstractmethod

from nehushtan.logger.NehushtanFileLogger import NehushtanFileLogger
from nehushtan.queue.NehushtanQueueTask import NehushtanQueueTask


class NehushtanQueueTaskDelegate:
    """
    Since 0.4.11
    """

    def __init__(self, config_dictionary: dict = None, logger: NehushtanFileLogger = None):
        if config_dictionary is None:
            config_dictionary = {}
        self.config_dictionary = config_dictionary

        if logger is None:
            logger = NehushtanFileLogger(f"{self.__class__.__name__}-Logger", )
        self.logger = logger

    @abstractmethod
    def when_to_execute_task(self, task: NehushtanQueueTask, pid: int):
        """
        SINCE 0.1.24 add parameter pid
        Note: Any exceptions should be caught inside.
        This works in WORKER - EMBEDED
        """
        pass

    @abstractmethod
    def when_task_raised_exception(self, task: NehushtanQueueTask, exception: Exception):
        """
        This works in WORKER - EMBEDED
        """
        pass

    @abstractmethod
    def when_task_executed(self, task: NehushtanQueueTask, pid: int):
        """
        SINCE 0.1.24 add parameter pid
        Note: Any exceptions should be caught inside.
        This works in WORKER - EMBEDED
        """
        try:
            if task.after_execute():
                self.logger.info('TASK FINISHED', {'task_id': task.get_task_reference(), 'pid': pid})
            else:
                self.logger.warning('TASK FINISHING UNWELL', {'task_id': task.get_task_reference(), 'pid': pid})
        except Exception as e:
            self.logger.exception(f'when_task_executed for task {task.get_task_reference()} on PID {pid}', e)



#  Copyright (c) 2020. Sinri Edogawa

from abc import abstractmethod, ABC

from nehushtan.mysql.MySQLKit import MySQLKit


class MySQLTableExistence(ABC):

    def __init__(self):
        """
        Since 0.2.17 Remove `self._mysql_kit`, it is not defined here anymore
        """
        pass

    @abstractmethod
    def get_mysql_kit(self) -> MySQLKit:
        pass

    @abstractmethod
    def mapping_table_name(self) -> str:
        pass

    @abstractmethod
    def mapping_schema_name(self) -> str:
        pass

    def get_table_expression(self) -> str:
        e = ''
        if self.mapping_schema_name() != '':
            e += f'`{self.mapping_schema_name()}`.'
        e += f'`{self.mapping_table_name()}`'
        return e



from nehushtan.httpd.NehushtanHTTPRequestController import NehushtanHTTPRequestController
from tests.httpd.TestRequestHandler import TestRequestHandler


class TestMQ(NehushtanHTTPRequestController):

    def product(self, queue_name: str, item: str):
        TestRequestHandler.mq.enqueue(item, queue_name)
        self._reply_with_ok("done")

    def consume(self, queue_name: str):
        x = TestRequestHandler.mq.dequeue(queue_name)
        if x is None:
            self._reply_with_ok({"mq_empty": True})
        else:
            self._reply_with_ok({"mq_empty": False, 'message': x})

    def statAll(self):
        x = TestRequestHandler.mq.stat_for_all()
        self._reply_with_ok({"mq_stat": x})



#  Copyright (c) 2020. Sinri Edogawa
from typing import Iterable

from nehushtan.mysql.MySQLCondition import MySQLCondition
from nehushtan.mysql.MySQLKit import MySQLKit
from nehushtan.mysql.MySQLSelectionMixin import MySQLSelectionMixin
from nehushtan.mysql.MySQLTableExistence import MySQLTableExistence


class MySQLTableSelection(MySQLSelectionMixin):
    """
    As of version 0.3.6, move some features to MySQLSelectionMixin
    """

    def __init__(self, model: MySQLTableExistence):
        super().__init__()
        self._model = model
        self._select_fields = []
        self._conditions = []
        self._group_by_fields = []
        self._sort_expression = ''
        self._limit = 0
        self._offset = 0
        self._use_indices = []
        self._force_indices = []
        self._ignore_indices = []
        self._for_update = False

    def get_mysql_kit(self) -> MySQLKit:
        return self._model.get_mysql_kit()

    def get_limit(self) -> int:
        return self._limit

    def set_limit(self, limit: int):
        self._limit = limit
        return self

    def get_offset(self) -> int:
        return self._offset

    def set_offset(self, offset: int):
        self._offset = offset
        return self

    def get_sort_expression(self):
        return self._sort_expression

    def set_sort_expression(self, sort_expression: str):
        self._sort_expression = sort_expression
        return self

    def add_select_field(self, field_expression: str, alias: str = ''):
        s = field_expression
        if len(alias) > 0:
            s += f' as {alias}'
        self._select_fields.append(s)
        return self

    def add_select_field_name_list(self, field_name_array: Iterable[str]):
        """

        :param field_name_array: ["F1","F2", ...] or ("F1","F2", ...)
        :return:
        """
        for item in field_name_array:
            self._select_fields.append(item)
        return self

    def add_condition(self, condition: MySQLCondition):
        self._conditions.append(condition)
        return self

    def add_conditions(self, conditions: Iterable[MySQLCondition]):
        """

        :param conditions: array of MySQLCondition
        :return:
        """
        for condition in conditions:
            if isinstance(condition, MySQLCondition):
                self._conditions.append(condition)
            # else just ignore
        return self

    def add_simple_conditions(self, equal_dict: dict):
        """

        :param equal_dict: {FIELD:VALUE,FIELD:VALUE_ARRAY}
        :return:
        """
        for k, v in equal_dict.items():
            if isinstance(v, (tuple, list)):
                self.add_condition(MySQLCondition.make_in_array(k, v))
            else:
                self.add_condition(MySQLCondition.make_equal(k, v))
        return self

    def set_group_by_fields(self, group_by_fields: list):
        self._group_by_fields = group_by_fields
        return self

    def use_index(self, index: str):
        self._use_indices.append(index)
        return self

    def force_index(self, index: str):
        self._force_indices.append(index)
        return self

    def ignore_index(self, index: str):
        self._ignore_indices.append(index)
        return self

    def set_for_update(self, value: bool):
        self._for_update = value

    def generate_sql(self) -> str:
        table = self._model.get_table_expression()

        fields = "*"
        if len(self._select_fields) > 0:
            fields = ",".join(self._select_fields)

        condition_sql = MySQLCondition.build_sql_component(self._conditions)

        indices = ''
        if len(self._use_indices) > 0:
            indices += " USE INDEX (" + ",".join(self._use_indices) + ") "
        if len(self._force_indices) > 0:
            indices += " FORCE INDEX (" + ",".join(self._force_indices) + ") "
        if len(self._ignore_indices) > 0:
            indices += " IGNORE INDEX (" + ",".join(self._force_indices) + ") "

        sql = f"SELECT {fields} FROM {table} {indices} WHERE {condition_sql} "

        if len(self._group_by_fields) > 0:
            sql += "GROUP BY " + ",".join(self._group_by_fields) + " "

        if self._sort_expression.strip() != '':
            sql += "ORDER BY " + self._sort_expression + " "

        if self._limit > 0:
            sql += f"LIMIT {self._limit} "
            if self._offset > 0:
                sql += f"OFFSET {self._offset} "

        if self._for_update:
            sql += " FOR UPDATE "

        return sql



import json

from nehushtan.httpd.NehushtanHTTPConstant import NehushtanHTTPConstant


class NehushtanHTTPResponseBuffer:
    """
    Since 0.4.0
    """

    BODY_DUMP_FORMAT_JSON = 'JSON'

    def __init__(self):
        self.http_code = 200
        self.http_code_message = 'OK'
        self.headers = {}
        self.__body = None
        self.body_dump_format = self.BODY_DUMP_FORMAT_JSON  # JSON, XML (planned only)
        self.encoding = None

    def set_encoding(self, encoding: str):
        self.encoding = encoding
        return self

    def set_http_code(self, code: int, message: str = None):
        self.http_code = code
        self.http_code_message = message
        return self

    def set_header(self, header_name: str, header_value: str):
        self.headers[header_name] = header_value
        return self

    def remove_header(self, header_name: str):
        del self.headers[header_name]
        return self

    def reset_body_as_string(self, init_string: str = ''):
        self.__body = init_string
        return self

    def append_string_to_body(self, appendix: str):
        if type(self.__body) is not str:
            raise TypeError('NehushtanHTTPResponseBuffer.append_string_to_body: body is not str')
        self.__body += appendix
        return self

    def get_body_as_string(self):
        if type(self.__body) is not str:
            if self.body_dump_format == self.BODY_DUMP_FORMAT_JSON:
                return self.get_body_as_json()
            raise TypeError('NehushtanHTTPResponseBuffer.get_body_as_string: body is not str')
        return self.__body

    def reset_body_as_dict(self, init_object: dict = None):
        if init_object is None:
            init_object = {}
        self.__body = init_object
        return self

    def get_body_as_dict(self) -> dict:
        if type(self.__body) is not dict:
            raise TypeError('NehushtanHTTPResponseBuffer.get_body_as_dict: body is not dict')
        return self.__body

    def reset_body_as_list(self, init_list: list = None):
        if init_list is None:
            init_list = []
        self.__body = init_list
        return self

    def get_body_as_list(self) -> list:
        if type(self.__body) is not list:
            raise TypeError('NehushtanHTTPResponseBuffer.get_body_as_list: body is not list')
        return self.__body

    def get_body_as_json(self) -> str:
        return json.dumps(self.__body)

    def refresh_content_length(self):
        s = self.get_body_as_string()
        if self.encoding is not None:
            s.encode(self.encoding)
        else:
            s.encode()
        self.set_header('Content-Length', '%i' % len(s))
        return self

    @staticmethod
    def make_for_redirect(url: str):
        return NehushtanHTTPResponseBuffer().set_http_code(302).set_header('Location', url)

    @staticmethod
    def make_for_json(anything):
        buffer = NehushtanHTTPResponseBuffer() \
            .set_http_code(200) \
            .set_header(NehushtanHTTPConstant.HEADER_CONTENT_TYPE, NehushtanHTTPConstant.HEADER_CONTENT_TYPE_VALUE_JSON)
        buffer.__body = anything
        buffer.refresh_content_length()
        return buffer



import json
import re

from nehushtan.helper.CommonHelper import CommonHelper
from nehushtan.httpd.NehushtanHTTPConstant import NehushtanHTTPConstant
from nehushtan.httpd.NehushtanHTTPRequestHandler import NehushtanHTTPRequestHandler
from nehushtan.httpd.NehushtanHTTPResponseBuffer import NehushtanHTTPResponseBuffer
from nehushtan.httpd.exceptions.NehushtanRequestParameterError import NehushtanRequestParameterError


class NehushtanHTTPRequestController:
    """
    Since 0.4.0
    """

    def __init__(self, http_handler: NehushtanHTTPRequestHandler):
        self.__http_handler = http_handler
        self.__response_buffer = NehushtanHTTPResponseBuffer()

    def _get_http_handler(self) -> NehushtanHTTPRequestHandler:
        return self.__http_handler

    def _get_response_buffer(self):
        return self.__response_buffer

    @staticmethod
    def __clean_paramter_with_regex(read_value, value_regex: str, key: str, key_type: str):
        if read_value is None:
            raise NehushtanRequestParameterError(
                key,
                key_type,
                NehushtanRequestParameterError.DESC_NOT_SET
            )
        if value_regex is not None:
            x = f'{read_value}'
            found = re.search(value_regex, x)
            if found is None:
                raise NehushtanRequestParameterError(
                    key,
                    key_type,
                    NehushtanRequestParameterError.DESC_FORMAT_ERROR
                )
        return read_value

    def _read_query(self, key: tuple, default=None):
        return CommonHelper.read_target(self._get_http_handler().parsed_query_dict, key, default)

    def _read_query_indispensably(self, key: tuple, value_regex: str = None):
        x = self._read_query(key)
        return self.__clean_paramter_with_regex(x, value_regex, f'{key}', NehushtanRequestParameterError.TYPE_QUERY)

    def _read_body(self, key: tuple, default=None):
        return CommonHelper.read_target(self._get_http_handler().parsed_body_data, key, default)

    def _read_body_indispensably(self, key: tuple, value_regex: str = None):
        x = self._read_body(key)
        return self.__clean_paramter_with_regex(x, value_regex, f'{key}', NehushtanRequestParameterError.TYPE_BODY)

    def _read_header(self, key: str, default: str = None):
        return self._get_http_handler().headers.get(key, default)

    def _read_cookie(self, key: str, default: str = None):
        return CommonHelper.read_target(self._get_http_handler().parsed_cookie_dict, (key,), default)

    def _respond_with_buffer(self):
        self._get_http_handler().send_response_with_buffer(self._get_response_buffer())

    def _reply_with_text(self, text: str, http_code: int = 200, encoding: str = None):
        buffer = self._get_response_buffer()
        if encoding is not None:
            buffer.encoding = encoding
        buffer.set_header(NehushtanHTTPConstant.HEADER_CONTENT_TYPE,
                          NehushtanHTTPConstant.HEADER_CONTENT_TYPE_VALUE_TEXT)
        buffer.set_http_code(http_code)
        buffer.reset_body_as_string(text)
        self._respond_with_buffer()

    def _reply_with_json(self, anything, http_code: int = 200, encoding: str = None):
        s = json.dumps(anything)
        buffer = self._get_response_buffer()
        if encoding is not None:
            buffer.encoding = encoding
        buffer.set_header(NehushtanHTTPConstant.HEADER_CONTENT_TYPE,
                          NehushtanHTTPConstant.HEADER_CONTENT_TYPE_VALUE_JSON)
        buffer.set_http_code(http_code)
        buffer.reset_body_as_string(s)
        self._respond_with_buffer()

    def _reply_with_ok(self, anything, encoding: str = None):
        self._reply_with_json({"code": NehushtanHTTPConstant.REPLY_CODE_OK, "data": anything}, 200, encoding)

    def _reply_with_fail(self, anything, encoding: str = None):
        self._reply_with_json({"code": NehushtanHTTPConstant.REPLY_CODE_FAIL, "data": anything}, 200, encoding)



import logging
import threading
import time
from random import random

from nehushtan.logger.NehushtanLogger import NehushtanLogger
from tests.config import test_log_store

shovel_logger = NehushtanLogger(
    logger_name="runner",
    handlers=(
        NehushtanLogger.make_stdout_handler(),
        NehushtanLogger.make_stderr_handler(),
        NehushtanLogger.make_fixed_file_handler(
            file_name=test_log_store + "/fixed.log"
        ),
        NehushtanLogger.make_timed_rotating_file_handler(
            file_name=test_log_store + "/rotating.log"
        )
    ),
    universal_log_level=logging.INFO,
    with_process_info=True,
    with_thread_info=True
)


def work(shovel_logger: NehushtanLogger, x):
    shovel_logger.info(f'I will sleep for {x} seconds...')
    time.sleep(x)
    shovel_logger.info('I woke up and die!')


thread_list = {}
for i in range(5):
    x = int(random() * 50)
    t = threading.Thread(target=work, args=(shovel_logger, x,))
    t.start()
    thread_list[i] = t

for i, t in thread_list.items():
    t.join()
    shovel_logger.info(f'joined {i}', threading.current_thread())

shovel_logger.info('awsl')



import time


class SearchCommandKit:
    """
    See https://datatracker.ietf.org/doc/html/rfc3501#section-6.4.4

    Since 0.4.6
    """

    def __init__(self, charset: str = None):
        self.__charset = charset
        self.__arguments = []
        if self.__charset is not None:
            self.__arguments.append('CHARSET')
            self.__arguments.append(self.__charset)

        self.__conditions = []

        self.__last_literal = None

    def get_charset(self):
        return self.__charset

    @staticmethod
    def formatted_date_from_time(x: float):
        """
        `x` might be get by `time.time()` and along with operations
        result would be `1-Jul-2021` alike
        """
        return time.strftime("%d-%b-%Y", time.localtime(x))

    def append_all(self):
        """
        All messages in the mailbox; the default initial key for ANDing.
        """
        self.__conditions.append('ALL')
        return self

    def append_answered(self):
        """
        Messages with the `\Answered` flag set.
        """
        self.__conditions.append('ANSWERED')
        return self

    def append_deleted(self):
        """
        Messages with the `\DELETED` flag set.
        """
        self.__conditions.append('DELETED')
        return self

    def append_draft(self):
        """
        Messages with the `\DRAFT` flag set.
        """
        self.__conditions.append('DRAFT')
        return self

    def append_flagged(self):
        """
        Messages with the `\FLAGGED` flag set.
        """
        self.__conditions.append('FLAGGED')
        return self

    def append_recent(self):
        """
        Messages with the `\RECENT` flag set.
        """
        self.__conditions.append('RECENT')
        return self

    def append_seen(self):
        """
        Messages with the `\SEEN` flag set.
        """
        self.__conditions.append('SEEN')
        return self

    def append_unanswered(self):
        """
        Messages that do not have the \Answered flag set.
        """
        self.__conditions.append('UNANSWERED')
        return self

    def append_undeleted(self):
        """
        Messages that do not have the \Deleted flag set.
        """
        self.__conditions.append('UNDELETED')
        return self

    def append_undraft(self):
        """
        Messages that do not have the \Draft flag set.
        """
        self.__conditions.append('UNDRAFT')
        return self

    def append_unflagged(self):
        """
        Messages that do not have the \Flagged flag set.
        """
        self.__conditions.append('UNFLAGGED')
        return self

    def append_unseen(self):
        """
        Messages that do not have the \Seen flag set.
        """
        self.__conditions.append('UNSEEN')
        return self

    def append_bcc(self, value: str):
        """
        Messages that contain the specified string in the envelope structure's BCC field.
        """
        self.__conditions.append(['BCC', value])
        return self

    def append_cc(self, value: str):
        """
        Messages that contain the specified string in the envelope structure's CC field.
        """
        self.__conditions.append(['CC', value])
        return self

    def append_from(self, value: str):
        """
        Messages that contain the specified string in the envelope structure's FROM field.
        """
        self.__conditions.append(['FROM', value])
        return self

    def append_to(self, value: str):
        """
        Messages that contain the specified string in the envelope structure's TO field.
        """
        self.__conditions.append(['TO', value])
        return self

    def append_before(self, before_date: str):
        """
        Messages whose internal date (disregarding time and timezone) is earlier than the specified date.
        """
        self.__conditions.append(['BEFORE', before_date])
        return self

    def append_on(self, on_date: str):
        """
        Messages whose internal date (disregarding time and timezone) is within the specified date.
        """
        self.__conditions.append(['ON', on_date])
        return self

    def append_since(self, since_date: str):
        """
        Messages whose internal date (disregarding time and timezone) is within or later than the specified date.
        """
        self.__conditions.append(['SINCE', since_date])
        return self

    def append_sent_before(self, before_date: str):
        """
        Messages whose [RFC-2822] Date: header (disregarding time and timezone) is earlier than the specified date.
        """
        self.__conditions.append(['SENTBEFORE', before_date])
        return self

    def append_sent_on(self, on_date: str):
        """
        Messages whose [RFC-2822] Date: header (disregarding time and timezone) is within the specified date.
        """
        self.__conditions.append(['SENTON', on_date])
        return self

    def append_sent_since(self, since_date: str):
        """
        Messages whose [RFC-2822] Date: header (disregarding time and timezone) is within the specified date.
        """
        self.__conditions.append(['SENTSINCE', since_date])
        return self

    def append_body_keyword(self, keyword: str):
        """
        Messages that contain the specified string in the body of the message.
        """
        self.__conditions.append(['BODY', keyword])
        return self

    def append_header_key_value(self, name: str, keyword: str):
        """
        Messages that have a header with the specified field-name (as defined in [RFC-2822])
            and that contains the specified string in the text of the header (what comes after the colon).
        If the string to search is zero-length,
            this matches all messages
            that have a header line with the specified field-name regardless of the contents.
        """
        self.__conditions.append(['HEADER', name, keyword])
        return self

    def append_flag_keyword(self, flag: str):
        """
        Messages with the specified keyword flag set.
        """
        self.__conditions.append(['KEYWORD', flag])
        return self

    def append_no_flag_keyword(self, flag: str):
        """
        Messages that do not have the specified keyword flag set.
        """
        self.__conditions.append(['UNKEYWORD', flag])
        return self

    def append_larger(self, octet_size: int):
        """
        Messages with an [RFC-2822] size larger than the specified number of octets.
        """
        self.__conditions.append(['LARGER', f'{octet_size}'])
        return self

    def append_smaller(self, octet_size: int):
        """
        Messages with an [RFC-2822] size smaller than the specified number of octets.
        """
        self.__conditions.append(['SMALLER', f'{octet_size}'])
        return self

    def append_new(self):
        """
        Messages that have the \Recent flag set but not the \Seen flag.
        This is functionally equivalent to "(RECENT UNSEEN)".
        """
        self.__conditions.append('NEW')
        return self

    def append_not_key(self, keyword: str):
        """
        Messages that do not match the specified search key.
        """
        self.__conditions.append(['NOT', keyword])
        return self

    def append_old(self):
        """
        Messages that do not have the \Recent flag set.
        This is functionally equivalent to "NOT RECENT" (as opposed to "NOT NEW").
        """
        self.__conditions.append('OLD')
        return self

    def append_or(self, v1: str, v2: str):
        """
        Messages that match either search key.
        """
        self.__conditions.append(['OR', v1, v2])
        return self

    def append_subject(self, keyword: str):
        """
        Messages that contain the specified string in the envelope structure's SUBJECT field.
        """
        self.__conditions.append(['SUBJECT', keyword])
        return self

    def append_text(self, keyword: str):
        """
        Messages that contain the specified string in the header or body of the message.
        """
        self.__conditions.append(['TEXT', keyword])
        return self

    def append_uid(self, uid: str):
        """
        Messages with unique identifiers corresponding to the specified unique identifier set.
        Sequence set ranges are permitted.
        """
        self.__conditions.append(['UID', uid])
        return self

    def build(self):
        t = []
        for x in self.__arguments:
            t.append(x)
        for condition in self.__conditions:
            if type(condition) is str:
                t.append(condition)
            elif type(condition) is list:
                for c in condition:
                    t.append(c)
                if condition[0] == 'SUBJECT' or condition[0] == 'TEXT':
                    self.__last_literal = condition[-1]
                    if self.__charset is not None:
                        self.__last_literal = self.__last_literal.encode(self.__charset)
                    t = t[:-1]
        return 'SEARCH', t, self.__last_literal



from threading import Thread
from typing import Dict


class SocketHandlerThreadManager:
    """
    Since 0.4.16
    """

    def __init__(self):
        self.__thread_map: Dict[str, Thread] = {}

    def register_thread(self, thread: Thread):
        self.__thread_map[thread.name] = thread

    def check_alive_thread_count(self):
        dead = []
        for thread_name, thread in self.__thread_map.items():
            if not thread.is_alive():
                dead.append(thread_name)
        for d in dead:
            del self.__thread_map[d]

        return len(self.__thread_map)



import unittest

from nehushtan.helper.CommonHelper import CommonHelper
from nehushtan.logger.NehushtanFileLogger import NehushtanFileLogger


class TestForCommonHelper(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        print('setUpClass', cls)

    @classmethod
    def tearDownClass(cls) -> None:
        print('tearDownClass', cls)

    def setUp(self):
        print('setUp')

    def tearDown(self):
        print('tearDown')

    def test_read_anything(self):
        target_1 = {
            'A': {
                "B": {
                    "C": "D",
                    "E": 0,
                    "F": False,
                    "G": True,
                    "H": None,
                },
                "I": [
                    "J",
                    {"K": "L"},
                    ["M"],
                ],
            },
        }
        # match
        self.assertEqual(
            target_1['A'],
            CommonHelper.read_target(target_1, ('A',))
        )
        self.assertEqual(
            target_1['A']['B'],
            CommonHelper.read_target(target_1, ('A', 'B')),
        )
        self.assertEqual(
            target_1['A']['B']['C'],
            CommonHelper.read_target(target_1, ('A', 'B', 'C'))
        )
        self.assertEqual(
            target_1['A']['B']['E'],
            CommonHelper.read_target(target_1, ('A', 'B', 'E'))
        )
        self.assertEqual(
            target_1['A']['B']['F'],
            CommonHelper.read_target(target_1, ('A', 'B', 'F'))
        )
        self.assertEqual(
            target_1['A']['B']['G'],
            CommonHelper.read_target(target_1, ('A', 'B', 'G'))
        )
        self.assertEqual(
            target_1['A']['B']['H'],
            CommonHelper.read_target(target_1, ('A', 'B', 'H'))
        )
        self.assertEqual(
            target_1['A']['I'],
            CommonHelper.read_target(target_1, ('A', 'I',))
        )
        self.assertEqual(
            target_1['A']['I'][0],
            CommonHelper.read_target(target_1, ('A', 'I', 0))
        )
        self.assertEqual(
            target_1['A']['I'][1],
            CommonHelper.read_target(target_1, ('A', 'I', 1))
        )
        self.assertEqual(
            target_1['A']['I'][1]['K'],
            CommonHelper.read_target(target_1, ('A', 'I', 1, 'K'))
        )
        self.assertEqual(
            target_1['A']['I'][2],
            CommonHelper.read_target(target_1, ('A', 'I', 2)),
        )
        # default
        self.assertEqual(
            'DEFAULT',
            CommonHelper.read_target(target_1, tuple(), 'DEFAULT')
        )
        self.assertEqual(
            'DEFAULT',
            CommonHelper.read_target(None, tuple(), 'DEFAULT')
        )
        self.assertEqual(
            'DEFAULT',
            CommonHelper.read_target(target_1, ('Z', 'Y', 'X',), 'DEFAULT')
        )
        self.assertEqual(
            'DEFAULT',
            CommonHelper.read_target(target_1, ('A', 'C', 0,), 'DEFAULT')
        )

    def test_write_dictionary(self):
        target = {}
        self.assertEqual(
            {'A': {'B': 'C'}},
            CommonHelper.write_dictionary(target, ('A', 'B',), 'C')
        )
        self.assertEqual(
            {'A': {'B': 'D'}},
            CommonHelper.write_dictionary(target, ('A', 'B',), 'D')
        )
        self.assertEqual(
            {'A': {'B': 'D', 'E': 'F'}, },
            CommonHelper.write_dictionary(target, ('A', 'E',), 'F')
        )

    def test_seek_class(self):
        self.assertIsInstance(
            NehushtanFileLogger(),
            CommonHelper.class_with_class_path('nehushtan.logger.NehushtanFileLogger', 'NehushtanFileLogger')
        )



import unittest

from nehushtan.cli.NehushtanArgumentParser import NehushtanArgumentParser


class TestForArgumentParser(unittest.TestCase):

    def test_for_parse(self):
        parser = NehushtanArgumentParser(['-a', 'aa', '--dd', 'ee', '--gg-hh', 'ii', '-i', '--jj', '--k-ll']) \
            .add_option('A', 'Option A', short='a') \
            .add_option('D', 'Option D', long='dd') \
            .add_option('G', 'Option G', prefix='gg') \
            .add_option('I', 'Flag I', short='i', is_flag=True) \
            .add_option('J', 'Flag J', long='jj', is_flag=True) \
            .add_option('K', 'Flag K', prefix='k', is_flag=True) \
            .add_option('L', 'Flag L', short='l', is_flag=True) \
            .parse()

        print(parser.get_usage_text())
        print(parser.get_raw_option_dict())

        parsed = parser.get_parsed_option_dict()
        expected = {'A': 'aa', 'D': 'ee', 'G': {'hh': 'ii'}, 'I': True, 'J': True, 'K': {"ll": True}}
        self.assertEqual(expected, parsed)

    def test_for_later_parse(self):
        try:
            NehushtanArgumentParser() \
                .add_option('A', 'arg A', short='a') \
                .add_option('D', 'arg D', short='a', long='dd') \
                .add_option('G', 'arg G', prefix='gg') \
                .set_arg_list(['-a', 'aa', '--dd', 'ee', '--gg-hh', 'ii', 'j', 'k']) \
                .parse()
        except RuntimeError as e:
            self.assertEqual('Short Key [-a] Already Existed', e.__str__())
        else:
            self.fail('Did not seek out the duplicated key')



import logging
import os

from nehushtan.logger.NehushtanLogger import NehushtanLogger
from tests.config import test_log_store


class FirstLoggerHolder:
    def __init__(self):
        self.logger = NehushtanLogger(
            'first',
            [
                NehushtanLogger.make_fixed_file_handler(os.path.join(test_log_store, 'first.log'))
            ]
        )


class SecondLoggerHolder:
    def __init__(self):
        self.logger = NehushtanLogger('first')


class ThirdLoggerHolder:
    def __init__(self):
        self.logger = NehushtanLogger(
            'first',
            [
                NehushtanLogger.make_stdout_handler(),
                NehushtanLogger.make_fixed_file_handler(os.path.join(test_log_store, 'third.log'))
            ]
        )


c1 = FirstLoggerHolder()
c1.logger.info('FIRST')

c2 = SecondLoggerHolder()
c2.logger.info('SECOND')

c1 = None
c2 = None

logging.getLogger('first').info('!', extra={"json_string": 'x'})
logging.getLogger('first').warning('!!', extra={"json_string": 'x'})
logging.getLogger('first').error('!!!', extra={"json_string": 'x'})

print(len(logging.getLogger('first').handlers))

c3 = ThirdLoggerHolder()
c3.logger.info('THIRD')

print(len(logging.getLogger('first').handlers))



import logging

from nehushtan.logger.NehushtanFileLogger import NehushtanFileLogger
from nehushtan.logger.NehushtanLogging import NehushtanLogging

if __name__ == '__main__':
    file_logger = NehushtanFileLogger('test', '/Users/leqee/code/nehushtan/log', log_level=logging.INFO,
                                      print_higher_than_this_level=NehushtanLogging.INFO,
                                      record_millisecond=True)

    file_logger.critical('C', {"A": "B"})
    file_logger.error('E', 111)
    file_logger.warning('W', 'dd')
    file_logger.info('I', file_logger)
    file_logger.debug('D')

    try:
        raise ValueError('A<>B')
    except Exception as error:
        file_logger.exception('X', error)

    args_json = file_logger.get_args_json_to_clone()
    print(args_json)
    cloned_logger = NehushtanFileLogger.build_instance_from_args_json(args_json)

    cloned_logger.critical('C', {"A": "B"})
    cloned_logger.error('E', 111)
    cloned_logger.warning('W', 'dd')
    cloned_logger.info('I', cloned_logger)
    cloned_logger.debug('D')

    try:
        raise ValueError('A<>B')
    except Exception as error:
        cloned_logger.exception('X', error)

    progress_logger = NehushtanFileLogger()
    total = 36
    for i in range(total):
        progress_logger.log_progress('TEST', i, total, desc=f'now done[{i}]')



class NehushtanLogging:
    CRITICAL = 50
    FATAL = 50
    ERROR = 40
    WARNING = 30
    WARN = 30
    NOTICE = 25
    INFO = 20
    DEBUG = 10
    NOTSET = 0

    @staticmethod
    def get_label_of_level(level: int):
        if level == NehushtanLogging.DEBUG:
            return 'DEBUG'
        elif level == NehushtanLogging.INFO:
            return 'INFO'
        elif level == NehushtanLogging.NOTICE:
            return 'NOTICE'
        elif level == NehushtanLogging.WARN or level == NehushtanLogging.WARNING:
            return 'WARNING'
        elif level == NehushtanLogging.ERROR:
            return 'ERROR'
        elif level == NehushtanLogging.CRITICAL or level == NehushtanLogging.FATAL:
            return 'CRITICAL'
        else:
            return 'NOTSET'

    @staticmethod
    def get_level_by_label(level_label: str):
        if level_label == 'DEBUG':
            return NehushtanLogging.DEBUG
        elif level_label == 'INFO':
            return NehushtanLogging.INFO
        elif level_label == 'NOTICE':
            return NehushtanLogging.NOTICE
        elif level_label == 'WARN' or level_label == 'WARNING':
            return NehushtanLogging.WARNING
        elif level_label == 'ERROR':
            return NehushtanLogging.ERROR
        elif level_label == 'CRITICAL' or level_label == 'FATAL':
            return NehushtanLogging.CRITICAL
        else:
            return NehushtanLogging.NOTSET



test_log_store = '/Users/leqee/code/nehushtan/log'

MYSQL_CONFIG = {
    'host': '172.16.1.52',
    'port': 3306,
    'user': 'ljni',
    'password': 'riIOTtIdBbT1',
    'db': 'sinri',
    'charset': 'utf8',
    'auto_commit': True,
}

IMAP_CONFIG = {
    # ...
}

SMTP_CONFIG = {
    # ...
}



import re
import sys


class NehushtanArgumentParser:
    """
    Since 0.2.19 Add Flag Support
    Support
    <python xxx.py> [-a B] [--bb C] [--dd-ee F --dd-gg H] [-i] [--jj] [--kk-ll --kk-mm]
    Suppose key is as argument name, the above would be parsed as
    {
        "a":"B"
        "bb":"C",
        "dd":{
            "ee":"F",
            "gg":"H",
        },
        "i":True,
        "jj":True,
        "kk":{
            "ll":True,
            "mm":True,
        }
    }
    """

    def __init__(self, sys_arg_list: list = None):
        """
        Since 0.2.19 sys_arg_list become optional
        :param sys_arg_list: a list of arguments, if omitted use `sys.argv[1:]`
        """
        if sys_arg_list is None:
            sys_arg_list = sys.argv[1:]

        self._sys_arg_list = sys_arg_list
        self._argument_dict = {}

        self._options_short = {}
        self._options_long = {}
        self._options_prefix = {}

        self._flags_short = {}
        self._flags_long = {}
        self._flags_prefix = {}

        self._result_dict = {}
        self._usage_dict = {}

    def set_arg_list(self, arg_list: list):
        """
        :param arg_list: Commonly `sys.argv[1:]`
        """
        self._sys_arg_list = arg_list
        return self

    def add_option(
            self,
            key: str, desc: str = '',
            short: str = '',
            long: str = '',
            prefix: str = '',
            is_flag=False
    ):
        """

        :param desc:
        :param key: Any Name
        :param short: a -> `-a`
        :param long: alpha -> `--alpha`
        :param prefix: alef -> `--alef-`
        :param is_flag: if its value follows existence only as boolean
        :return:
        """
        if short == '' and long == '' and prefix == '':
            return self

        usage = {
            'key': key,
            'desc': desc,
        }

        if short != '':
            argument_key = f'-{short}'
            self.__assert_key_unique(argument_key, 'short')
            if is_flag:
                self._flags_short[argument_key] = key
                usage['short'] = f'[-{short}]'
            else:
                self._options_short[argument_key] = key
                usage['short'] = f'-{short} [VALUE]'
        if long != '':
            argument_key = f'--{long}'
            self.__assert_key_unique(argument_key, 'long')
            if is_flag:
                self._flags_long[argument_key] = key
                usage['long'] = f'[--{long}]'
            else:
                self._options_long[argument_key] = key
                usage['long'] = f'--{long} [VALUE]'
        if prefix != '':
            argument_key = f'--{prefix}-'
            self.__assert_key_unique(argument_key, 'prefix')
            if is_flag:
                self._flags_prefix[argument_key] = key
                usage['prefix'] = f'[--{prefix}-[KEY]]'
            else:
                self._options_prefix[argument_key] = key
                usage['prefix'] = f'--{prefix}-[KEY] [VALUE]'

        self._usage_dict[key] = usage

        return self

    def __assert_key_unique(self, argument_key: str, key_type: str):
        if key_type == 'short':
            if self._flags_short.get(argument_key) is not None or self._options_short.get(argument_key) is not None:
                raise RuntimeError(f'Short Key [{argument_key}] Already Existed')
        elif key_type == 'long':
            if self._flags_long.get(argument_key) is not None or self._options_long.get(argument_key) is not None:
                raise RuntimeError(f'Long Key [{argument_key}] Already Existed')
        elif key_type == 'prefix':
            if self._flags_prefix.get(argument_key) is not None or self._options_prefix.get(argument_key) is not None:
                raise RuntimeError(f'Prefix Key [{argument_key}] Already Existed')

    def parse(self):
        self._result_dict = {}

        current_key = None
        self._argument_dict = {}
        for item in self._sys_arg_list:
            if current_key is None:
                # check if current item is a flag key
                if self.__parse_flag_key(item) is True:
                    self._argument_dict[item] = True
                    continue

                current_key = item
            else:
                self._argument_dict[current_key] = item
                current_key = None

        for (key, value) in self._argument_dict.items():
            self.__parse_option_pair(key, value)

        return self

    def __parse_flag_key(self, key):
        if self._flags_short.get(key) is not None:
            self._result_dict[self._flags_short[key]] = True
            return True
        elif self._flags_long.get(key) is not None:
            self._result_dict[self._flags_long[key]] = True
            return True
        else:
            for prefix, mapping_key in self._flags_prefix.items():
                matches = re.match(f'^{prefix}([A-Za-z0-9._-]+)$', key)
                if matches is not None:
                    if self._result_dict.get(mapping_key) is None:
                        self._result_dict[mapping_key] = {}
                    self._result_dict[mapping_key][matches[1]] = True
                    return True
        # not matched any
        return False

    def __parse_option_pair(self, key: str, value):
        if self._options_short.get(key) is not None:
            self._result_dict[self._options_short[key]] = value
        elif self._options_long.get(key) is not None:
            self._result_dict[self._options_long[key]] = value
        else:
            for prefix, mapping_key in self._options_prefix.items():
                matches = re.match(f'^{prefix}([A-Za-z0-9._-]+)$', key)
                if matches is not None:
                    if self._result_dict.get(mapping_key) is None:
                        self._result_dict[mapping_key] = {}
                    self._result_dict[mapping_key][matches[1]] = value
                    break

    def get_raw_option_dict(self):
        return self._argument_dict

    def get_parsed_option_dict(self):
        return self._result_dict

    def get_usage_text(self):
        usage_text = ""
        for key, value in self._usage_dict.items():
            options = []
            if len(value.get('short', '')) > 0:
                options.append(value.get('short'))
            if len(value.get('long', '')) > 0:
                options.append(value.get('long'))
            if len(value.get('prefix', '')) > 0:
                options.append(value.get('prefix'))

            usage_text += f"{value['key']}" + "\t\t"
            usage_text += ", ".join(options)
            if len(value.get('desc', '')) > 0:
                usage_text += f"\n{value['desc']}"
            usage_text += "\n"

        return usage_text



#  Copyright (c) 2020. Sinri Edogawa

MYSQL_QUERY_STATUS_INIT = "INIT"
MYSQL_QUERY_STATUS_QUERIED = "QUERIED"
MYSQL_QUERY_STATUS_EXECUTED = "EXECUTED"
MYSQL_QUERY_STATUS_STREAMING = "STREAMING"
MYSQL_QUERY_STATUS_STREAMED = "STREAMED"
MYSQL_QUERY_STATUS_ERROR = "ERROR"

MYSQL_QUERY_ROW_TYPE_DICT = 'DICT'
MYSQL_QUERY_ROW_TYPE_TUPLE = 'TUPLE'
MYSQL_QUERY_ROW_TYPE_UNKNOWN = 'UNKNOWN'



from nehushtan.httpd.exceptions.NehushtanHTTPError import NehushtanHTTPError


class NehushtanRequestDeniedByFilterError(NehushtanHTTPError):
    """
    Since 0.4.0
    When a Filter denies a request
    """

    def __init__(self, filter_name: str, error_message: str, http_code: int):
        super(NehushtanRequestDeniedByFilterError, self).__init__(
            f'Filter [{filter_name}] Denied Request: {error_message}',
            http_code
        )



from nehushtan.mysql.MySQLKit import MySQLKit
from nehushtan.mysql.MySQLViewMixin import MySQLViewMixin


class MySQLAnyView(MySQLViewMixin):
    """
    Since 0.2.17
    """

    def __init__(self, mysql_kit: MySQLKit, view_name: str, schema_name: str = ''):
        super().__init__()
        self.__mysql_kit = mysql_kit
        self.__view_name = view_name
        self.__schema_name = schema_name

    def __del__(self):
        """
        Since 0.4.10
        """
        self.__mysql_kit = None

    def get_mysql_kit(self) -> MySQLKit:
        return self.__mysql_kit

    def mapping_table_name(self) -> str:
        return self.__view_name

    def mapping_schema_name(self) -> str:
        return self.__schema_name



#  Copyright (c) 2020. Sinri Edogawa

from nehushtan.mysql.MySQLKit import MySQLKit
from nehushtan.mysql.MySQLTableMixin import MySQLTableMixin


class MySQLAnyTable(MySQLTableMixin):

    def __init__(self, mysql_kit: MySQLKit, table_name: str, schema_name: str = ''):
        super().__init__()
        self._mysql_kit = mysql_kit
        self._table_name = table_name
        self._schema_name = schema_name

    def __del__(self):
        """
        Since 0.4.10
        """
        self._mysql_kit = None

    def get_mysql_kit(self) -> MySQLKit:
        return self._mysql_kit

    def mapping_table_name(self) -> str:
        return self._table_name

    def mapping_schema_name(self) -> str:
        return self._schema_name



from abc import abstractmethod
from typing import List


class NehushtanQueueTask:
    def __init__(self):
        self.ready_to_execute: bool = False
        self.ready_to_finish: bool = False
        self.done: bool = False
        self.execute_feedback: str = 'Not Executed Yet'
        self.execute_result = None

    @abstractmethod
    def get_task_reference(self):
        pass

    @abstractmethod
    def get_task_type(self):
        pass

    def before_execute(self):
        """
        This works in MASTER
        """
        self.ready_to_execute = True
        return self.ready_to_execute

    @abstractmethod
    def execute(self):
        """
        This works in WORKER - EMBEDED

        Need to override and fulfill
        self.done: bool
        self.execute_feedback: str
        self.execute_result: Any
        """
        pass

    def after_execute(self):
        """
        This works in WORKER - EMBEDED
        """
        self.ready_to_finish = True
        return self.ready_to_finish

    def get_lock_list(self) -> List[str]:
        """
        Array of Lock Names
        """
        return []

    def is_exclusive(self) -> bool:
        """
        Determine if this task should be run in Exclusive Mode,
        which would wait for other all tasks stopped before starts,
        and never run other tasks until it ends.

        This works in MASTER
        """
        return False

    def check_if_locked(self) -> bool:
        """
        Locks should be checked here
        Should be called by Delegate::checkNextTaskImplement
        """
        return False



import re
from typing import Sequence, Tuple, Union

from nehushtan.httpd.NehushtanHTTPRoute import NehushtanHTTPRoute


class NehushtanHTTPRouteWithRegexArgs(NehushtanHTTPRoute):
    """
    Since 0.4.0
    """

    def __init__(
            self,
            path_template: str,
            controller_target: Tuple[Union[type, str], str],
            filter_list: Sequence[Union[type, str]] = None,
            method_options: Sequence[str] = None
    ):
        """
        self.path_template is expected as '/FIXED/(\\w+)/...'
        Since 0.4.1, the `^` and `$` are not needed; special keyword designed:
            {int} -> r'(\d+)'
            {float} -> r'(\d\.?\d*)'
            {str} -> r'([^/]+)'
        """
        super().__init__()
        # self.regex_for_path = path_template.replace('{int}', r'(\d+)') \
        #     .replace('{float}', r'(\d\.?\d*)') \
        #     .replace('{str}', r'([^/]+)')
        # if self.regex_for_path[0] != '^':
        #     self.regex_for_path = '^' + path_template
        # if self.regex_for_path[-1] != '$':
        #     self.regex_for_path = path_template + '$'
        self.path_template = path_template
        self.method_options = method_options
        self.filter_list = filter_list if filter_list is not None else []
        self.controller_target = controller_target

    def match_request(self, method: str, path: str) -> bool:
        if self.method_options is not None:
            if not self.method_options.__contains__(method):
                return False

        path_components = path.split('/')
        path_template_components = self.path_template.split('/')

        matched_arguments = []

        if len(path_components) < len(path_template_components):
            return False

        for i in range(len(path_components)):
            path_component = path_components[i]

            if i >= len(path_template_components):
                matched_arguments.append(path_component)
                continue

            path_template_component = path_template_components[i]
            if path_component == path_template_component:
                # static
                continue
            elif path_template_component == '{int}' and re.search(r'^\d+$', path_component):
                matched_arguments.append(int(path_component))
            elif path_template_component == '{float}' and re.search(r'^\d+\.?\d*$', path_component):
                matched_arguments.append(float(path_component))
            elif path_template_component == '{str}':
                matched_arguments.append(path_component)
            else:
                return False

        self.matched_arguments = matched_arguments

        # # old
        # result = re.search(self.regex_for_path, path)
        # if result is None:
        #     return False
        #
        # self.matched_arguments = result.groups()

        return True

    def get_filter_list(self) -> Sequence[Union[type, str]]:
        return self.filter_list

    def get_controller_target(self) -> Tuple[Union[type, str], str]:
        return self.controller_target

    def __str__(self):
        method = "|".join(self.method_options) if self.method_options is not None else 'ANY'
        return f'NehushtanHTTPRouteWithRegexArgs [{method}] {self.path_template}'



from pymysql import MySQLError, InterfaceError

from nehushtan.mysql.MySQLKit import MySQLKit
from nehushtan.mysql.MySQLKitConfig import MySQLKitConfig
from tests.config import MYSQL_CONFIG

db = MySQLKit(MySQLKitConfig(MYSQL_CONFIG))

cursor = db.get_raw_connection().cursor()
cursor.execute("select id,data from a limit 2")
all = cursor.fetchall()
print('first', all)

db.get_raw_connection().close()
try:
    cursor.execute("select id,data from a limit 2")
    all = cursor.fetchall()
    print('second', all)
except InterfaceError as ie:
    print(ie, ie.__cause__)
except MySQLError as e:
    print(e)
    print(e.__class__, e.__class__.__doc__)



from nehushtan.logger.NehushtanLogger import NehushtanLogger
from nehushtan.mysql.MySQLAnyTable import MySQLAnyTable
from nehushtan.mysql.MySQLKit import MySQLKit
from nehushtan.mysql.MySQLKitConfig import MySQLKitConfig
from tests.config import MYSQL_CONFIG

logger = NehushtanLogger('test', [NehushtanLogger.make_stdout_handler()])

table = MySQLAnyTable(
    MySQLKit(MySQLKitConfig(MYSQL_CONFIG)),
    'b'
)

table.truncate()

result = table.insert_one_row(
    {
        'name': 'one',
        'value': 1,
        'price': 1.1,
        'meta': None,
        'update_time': table.now(),
    }
)

logger.info(
    'insert one row',
    {
        'status': result.get_status(),
        'last_insert_id': result.get_last_inserted_id(),
        'afx': result.get_affected_rows(),
    }
)

dict_rows = []
for i in range(10000):
    dict_rows.append({
        'name': f'batch-name-{i}',
        'value': i,
        'price': i * 0.3,
        'meta': None,
        'update_time': table.now(),
    })

logger.info('insert_many_rows_with_dicts for 10000 rows start')
result = table.insert_many_rows_with_dicts(dict_rows)
logger.info(
    'insert_many_rows_with_dicts for 10000 rows end',
    {
        'status': result.get_status(),
        'last_insert_id': result.get_last_inserted_id(),
        'afx': result.get_affected_rows(),
        'error': result.get_error(),
    }
)

dict_rows = []
for i in range(10000):
    dict_rows.append({
        'name': f'replace-name-{i}',
        'value': i,
        'price': i * 0.4,
        'meta': 'fuck',
        'update_time': table.now(),
    })

logger.info('replace_many_rows_with_dicts for 10000 rows start')
result = table.replace_many_rows_with_dicts(dict_rows)
logger.info(
    'replace_many_rows_with_dicts for 10000 rows end',
    {
        'status': result.get_status(),
        'last_insert_id': result.get_last_inserted_id(),
        'afx': result.get_affected_rows(),
        'error': result.get_error(),
    }
)



import re
from email import policy
from email.message import EmailMessage, Message
from email.parser import BytesParser
from typing import Union

from nehushtan.mail.rfc822.NehushtanMessagePart import NehushtanMessagePart


class NehushtanEmailMessage:
    """
    Since 0.4.8
    """

    def __init__(self, msg: Union[EmailMessage, Message]):
        self.__email_message = msg
        self.__part = NehushtanMessagePart(self.__email_message, None)

    def get_email_message(self):
        return self.__email_message

    def get_part(self):
        return self.__part

    def read_field_from(self):
        return self.__email_message.get('From', '')

    def read_field_to(self):
        return self.__email_message.get('To', '')
        # x=self.__email_message.get('To','')
        # return NehushtanEmailMessage.parse_mail_address_line(x)

    def read_field_cc(self):
        return self.__email_message.get('cc', '')

    def read_field_bcc(self):
        return self.__email_message.get('bcc', '')

    def read_field_subject(self):
        return self.__email_message.get('Subject', '')

    def read_field_date(self):
        return self.__email_message.get('Date', '')

    @staticmethod
    def parse_bytes(text: bytes):
        # <class 'email.message.EmailMessage'>
        msg = BytesParser(policy=policy.default).parsebytes(text)
        return NehushtanEmailMessage(msg)

    @staticmethod
    def parse_mail_address_line(x: str):
        # print('parse_mail_address_line <-', x)

        # a1@b.c, a2@b.c
        # A1 <a1@b.c>, a2@b.c
        # A1 <a1@b.c>, A2 <a2@b.c>

        email_rule_express = r'([A-Za-z0-9.-]+@[A-Za-z0-9.-]+)|((\S*)\s*<([A-Za-z0-9.-]+@[A-Za-z0-9.-]+)>)'
        address_pieces = re.findall(email_rule_express, x)

        list_of_mail_address_tuple = []
        for address_piece in address_pieces:
            address = None
            name = None
            if address_piece[0]:
                address = address_piece[0]
                name = None
            elif address_piece[2] and address_piece[3]:
                address = address_piece[3]
                name = address_piece[2]
                # EncodedWordsKit.decode_string_following_rfc2047(name)

            list_of_mail_address_tuple.append((address, name))
        return list_of_mail_address_tuple

    @staticmethod
    def extract_email_address(x: str):
        """
        Since 0.4.9
        """
        email_rule_express = r'[A-Za-z0-9.-]+@[A-Za-z0-9.-]+'
        return re.findall(email_rule_express, x)



import re
from typing import Sequence, Tuple, Union

from nehushtan.httpd.NehushtanHTTPRoute import NehushtanHTTPRoute


class NehushtanHTTPRouteWithRegexKwargs(NehushtanHTTPRoute):
    """
    Since 0.4.0
    """

    def __init__(
            self,
            path_template: str,
            controller_target: Tuple[Union[type, str], str],
            filter_list: Sequence[Union[type, str]] = None,
            method_options: Sequence[str] = None
    ):
        """
        self.path_template is expected as /FIXED/{NAMED}/...
        Since 0.4.1 `{NAMED}` could be extended as `{NAMED:int|float|str}`
        """

        super().__init__()
        self.filter_list = filter_list if filter_list is not None else []
        self.controller_target = controller_target
        self.method_options = method_options

        pairs = re.findall(r'{([A-Za-z0-9_]+):?([a-z]+)?}', path_template)
        self.kwargs_type_dict = {}
        for kwargs_name, kwargs_type in pairs:
            self.kwargs_type_dict[kwargs_name] = kwargs_type

        # print(path_template)
        x = re.sub(r'{([A-Za-z0-9_]+):int}', r'(?P<\1>\\d+)', path_template)
        x = re.sub(r'{([A-Za-z0-9_]+):float}', r'(?P<\1>\\d+\.?\\d+)', x)
        x = re.sub(r'{([A-Za-z0-9_]+)(:str)?}', r'(?P<\1>[^/]+)', x)
        # print(x)

        self.named_path_template = f'^{x}$'

    def match_request(self, method: str, path: str) -> bool:
        if self.method_options is not None:
            if not self.method_options.__contains__(method):
                return False

        result = re.search(self.named_path_template, path)
        if result is None:
            return False

        # self.matched_arguments = result.groups()
        # self.matched_keyed_arguments = result.groupdict()

        self.matched_keyed_arguments = {}
        for k, v in result.groupdict().items():
            t = self.kwargs_type_dict.get(k, 'str')
            if t == 'int':
                self.matched_keyed_arguments[k] = int(v)
            elif t == 'float':
                self.matched_keyed_arguments[k] = float(v)
            else:
                self.matched_keyed_arguments[k] = v

        return True

    def get_filter_list(self) -> Sequence[Union[type, str]]:
        return self.filter_list

    def get_controller_target(self) -> Tuple[Union[type, str], str]:
        return self.controller_target

    def __str__(self):
        method = "|".join(self.method_options) if self.method_options is not None else 'ANY'
        return f'NehushtanHTTPRouteWithRegexKwargs [{method}] {self.named_path_template}'



from nehushtan.httpd.NehushtanHTTPRequestController import NehushtanHTTPRequestController


class TestProcessChainA(NehushtanHTTPRequestController):
    def read(self, index: int):
        # self.get_http_handler().log_message('emmm')
        # self._get_http_handler().send_response(200)
        # self._get_http_handler().send_header('content-type', 'text/plain')
        # self._get_http_handler().end_headers()
        # self._get_http_handler().wfile.write(f'READ INDEX [{index}]'.encode())
        self._reply_with_ok({'index': index})

    def step1(self, index: str):
        self._get_http_handler().get_process_chain_share_data_dict()['index'] = int(index)

    def step2(self):
        x = self._get_http_handler().get_process_chain_share_data_dict().get('index', 0)
        self._get_http_handler().get_process_chain_share_data_dict()['index'] = (x + 1)



import os
import time
from multiprocessing import Process

from nehushtan.queue.NehushtanQueueDelegate import NehushtanQueueDelegate
from nehushtan.queue.NehushtanQueueTask import NehushtanQueueTask
from nehushtan.queue.NehushtanQueueTaskDelegate import NehushtanQueueTaskDelegate
from nehushtan.queue.situation.NoNextTaskSituation import NoNextTaskSituation


class NehushtanQueue:
    def __init__(self, delegate: NehushtanQueueDelegate, task_delegate: NehushtanQueueTaskDelegate):
        """
        Since 0.4.11 add task_delegate
        """
        self.delegate = delegate
        self.task_delegate = task_delegate
        self.current_workers = {}
        self._is_running_exclusive_task = False

    def get_logger(self):
        return self.delegate.logger

    def get_pool_capacity(self) -> int:
        """
        How many processes could run in the same time.
        Decided by the delegate
        """
        if self._is_running_exclusive_task is True:
            return 1

        return self.delegate.read_config_of_delegate((NehushtanQueueDelegate.CONFIG_KEY_POOL_CAPACITY,), 1)

    def _scan_workers(self):
        """
        Scan registered workers, and remove dead ones from dictionary.
        Finally return the total running processes count.
        """
        if len(self.current_workers.items()) > 0:
            delete_keys = []
            for task_id, task_process in self.current_workers.items():
                if not task_process.is_alive():
                    task_process_pid = task_process.pid
                    task_process_exit_code = task_process.exitcode
                    self.get_logger().warning(f'Worker Process for task [{task_id}] Dead', {'pid': task_process_pid})
                    self.delegate.when_worker_process_confirmed_dead(
                        task_id,
                        {'pid': task_process_pid, 'exitcode': task_process_exit_code}
                    )
                    delete_keys.append(task_id)
            for delete_key in delete_keys:
                del self.current_workers[delete_key]
                self.get_logger().debug('remove removable finished tasks', {'task_id': delete_key})

        return len(self.current_workers.items())

    def _terminate_worker_process_of_task(self, task_reference):
        p = self.current_workers.get(task_reference)
        if p:

            # 终止进程
            # https://docs.python.org/zh-cn/3/library/multiprocessing.html#multiprocessing.Process.terminate
            #
            #   在Unix上
            #       这是使用 SIGTERM 信号完成的；
            #   在Windows上
            #       使用 TerminateProcess() 。请注意，不会执行退出处理程序和finally子句等。
            #       这个函数可以用来终止或者说杀死一个进程，它不会留给进程及其所有线程清理的时间，
            #       系统会马上终止(杀死)这个进程的所有线程，致使进程终止。
            #
            # 请注意，进程的后代进程将不会被终止 —— 它们将简单地变成孤立的。
            p.terminate()

            self.get_logger().info(
                'NehushtanQueue._terminate_worker_process_of_task executed',
                {'task': task_reference, 'pid': p.pid, 'exitcode': p.exitcode, 'is_alive': p.is_alive()}
            )
            self.delegate.when_killed_worker_process(task_reference, worker_pid=p.pid)
            self.register_news(
                'when_killed_worker_process',
                f'Sent signal TERMINATE(15) to worker for task {task_reference}, target pid {p.pid}'
            )
        else:
            self.get_logger().warning(
                'NehushtanQueue._terminate_worker_process_of_task target task not found amongst current workers',
                {'task': task_reference}
            )
            self.delegate.when_killed_worker_process(task_reference, not_found=True)

    def _loop_maintain(self, wait_till_processes_become_less_than: int = 0):
        self._scan_workers()

        if self.delegate.handle_command_queue() > 0:
            # Since 0.2.18 the news only be sent when any command(s) done
            self.register_news('handle_command_queue', 'Handled Command Queue')

        kill_list = self.delegate.should_kill_any_worker_processes()
        if len(kill_list) > 0:
            for kill_task_id in kill_list:
                self._terminate_worker_process_of_task(kill_task_id)

        total_alive_current_workers = self._scan_workers()

        if wait_till_processes_become_less_than > 0:
            # need to do some wait
            if total_alive_current_workers >= wait_till_processes_become_less_than:
                time.sleep(2)
                return self._loop_maintain(wait_till_processes_become_less_than)
        else:
            return total_alive_current_workers

    def register_news(self, title: str, content: str):
        """
        Since 0.2.9
        """
        worker_count = len(self.current_workers.items())
        busy_rate = 1.0 * worker_count / self.get_pool_capacity()
        self.delegate.register_queue_news(title, content, worker_count, busy_rate)

    def loop(self):
        self.delegate.when_loop_starts()
        self.register_news('when_loop_starts', 'Loop Started')

        while True:
            try:
                # 1. check current workers status
                total_current_workers = len(self.current_workers.items())
                self.get_logger().info('Current workers totally count', {'total': total_current_workers})

                # 2. handle command queue and kill list
                self._loop_maintain()

                # 3. check elihu command
                self.delegate.read_latest_command()
                if self.delegate.should_terminate():
                    if self.delegate.should_wait_for_all_workers_before_terminating():
                        self.register_news('terminating',
                                           'Waiting for all running tasks finishing, then terminate loop')
                        self._loop_maintain(1)
                    break

                # 4. check queue if new task there
                if not self.delegate.is_runnable():
                    self.register_news('should_not_run', 'Loop is ordered to sleep for a while')
                    self.delegate.when_loop_should_not_run()
                    continue

                total_current_workers = len(self.current_workers.items())
                if total_current_workers >= self.get_pool_capacity():
                    self.register_news('all_workers_busy', 'All workers are busy, wait till anyone become available')
                    self._loop_maintain(self.get_pool_capacity())

                    if self._is_running_exclusive_task and len(self.current_workers.items()) <= 0:
                        self._is_running_exclusive_task = False
                        self.register_news('exit_exclusive_mode', 'Exit exclusive mode as no task running now')

                try:
                    self.delegate.before_seeking_next_tasks()
                    task_candidates = self.delegate.check_next_task_candidates()

                    newly_locked = {}
                    for task_candidate in task_candidates:

                        # Since 0.3.1
                        # check locks and maintain a newly locked dict to avoid trying start up for known new locks
                        this_lock_list = task_candidate.get_lock_list()

                        already_locked = False
                        for x in this_lock_list:
                            if newly_locked.get(x, False) is True:
                                already_locked = True
                                break
                        if already_locked:
                            self.get_logger().debug(
                                f'Task Candidate {task_candidate.get_task_reference()} '
                                f'locked by newly running tasks, passover'
                            )
                            continue
                        else:
                            for x in this_lock_list:
                                newly_locked[x] = True

                            self.__run_one_task(task_candidate)
                            # here is a problem:
                            # when __run_one_task return False, the task_candidate actually did not start up,
                            # the locks are really not locked newly

                            if self._is_running_exclusive_task:
                                break
                            if self.get_pool_capacity() <= len(self.current_workers.items()):
                                break

                except NoNextTaskSituation:
                    self.register_news('when_no_task_to_do', 'There is no task in queue able to deal now')
                    self.delegate.when_no_task_to_do()
                    continue

            except Exception as loop_round_error:
                self.register_news('when_loop_reports_error', f'Loop meets error: {loop_round_error}')
                self.delegate.when_loop_reports_error(loop_round_error.__str__(), loop_round_error)

        self.register_news('loop_terminating', 'Loop is about to terminate')
        self.delegate.when_loop_terminates()
        self.register_news('when_loop_terminated', 'Loop terminated')

    def __run_one_task(self, task: NehushtanQueueTask) -> bool:
        """
        Since 0.3.0
        When a found task:
            successfully genereate a process to execute, return True.
            not executable, return False.
        """
        task_id = task.get_task_reference()

        if task.is_exclusive():
            # now daemon should wait for the other tasks to be over
            self.register_news('before_enter_exclusive_mode',
                               f'Task {task_id} is exclusive, waiting for all running workers finish')
            self._loop_maintain(1)
            self.register_news('enter_exclusive_mode',
                               f'Enter exclusive mode for task {task_id} as all other workers finished')
            self._is_running_exclusive_task = True

        if not task.before_execute():
            self.register_news('when_task_not_executable', f'Task {task_id} is not executable')
            self.delegate.when_task_not_executable(task)

            if self._is_running_exclusive_task:
                self._is_running_exclusive_task = False
                self.register_news('exit_exclusive_mode',
                                   f'Exit exclusive mode as task {task_id} is not executable')

            return False

        p = Process(target=NehushtanQueue.embedded_task_execute, args=(task, self.task_delegate))
        p.start()
        self.current_workers[task.get_task_reference()] = p

        self.register_news(
            'when_worker_process_created',
            f'Task [{task.get_task_reference()}] -> Process {p.name} PID is [{p.pid}]'
        )
        self.delegate.when_worker_process_created(
            task.get_task_reference(),
            f'Task [{task.get_task_reference()}] -> Process {p.name} PID is [{p.pid}]'
        )

        return True

    @staticmethod
    def embedded_task_execute(embedded_task: NehushtanQueueTask, task_delegate: NehushtanQueueTaskDelegate):
        """
        This works in WORKER
        Since 0.4.11 use NehushtanQueueTaskDelegate
        """
        try:
            """
            Since 0.4.5 Add try block to catch outer_exception
            """
            task_delegate.when_to_execute_task(embedded_task, os.getpid())
            try:
                embedded_task.execute()
            except Exception as e:
                task_delegate.when_task_raised_exception(embedded_task, e)
            task_delegate.when_task_executed(embedded_task, os.getpid())

        except Exception as outer_exception:
            task_delegate.logger.exception(
                'In embedded_task_execute exception occurs outside the embedded_task '
                f'[{embedded_task.get_task_reference()}] on Process {os.getpid()}',
                outer_exception
            )



import logging

from nehushtan.logger.NehushtanLogger import NehushtanLogger
from tests.config import test_log_store


class TestFilter(logging.Filter):
    def filter(self, record):
        print('filtering', record)
        print(f"record.levelno={record.levelno}")
        print(f"record.name={record.name}")
        print(f"record.thread={record.thread}")
        print(f"record.threadName={record.threadName}")
        print(f"record.message={record.message}")
        print(f"record.args={record.args}")
        print(f"record.asctime={record.asctime}")
        print(f"record.created={record.created}")
        print(f"record.exc_info={record.exc_info}")
        print(f"record.exc_text={record.exc_text}")
        print(f"record.filename={record.filename}")
        print(f"record.funcName={record.funcName}")
        print(f"record.lineno={record.lineno}")
        print(f"record.msecs={record.msecs}")
        print(f"record.msg={record.msg}")
        print(f"record.pathname={record.pathname}")
        print(f"record.process={record.process}")
        print(f"record.processName={record.processName}")
        print(f"record.threadName={record.threadName}")
        print(f"record.relativeCreated={record.relativeCreated}")
        print(f"record.__dict__={record.__dict__}")
        return True


filtering_handler = NehushtanLogger.make_fixed_file_handler(
    file_name=test_log_store + "/filtering.log",
)
filtering_handler.addFilter(TestFilter())

shovel_logger = NehushtanLogger(
    logger_name="runner",
    handlers=(
        NehushtanLogger.make_stdout_handler(),
        NehushtanLogger.make_stderr_handler(),
        NehushtanLogger.make_fixed_file_handler(
            file_name=test_log_store + "/fixed.log"
        ),
        NehushtanLogger.make_timed_rotating_file_handler(
            file_name=test_log_store + "/rotating.log"
        ),
        filtering_handler
    ),
    universal_log_level=logging.INFO
)

shovel_logger.debug('世界非常和平')
shovel_logger.info('世界还算凑合', "Guhehe")
shovel_logger.warning('世界有点问题', {"log_path": test_log_store})
shovel_logger.error('世界打起来了', shovel_logger)
shovel_logger.critical('世界马上灭亡', [3, 2, 1, {0: [1, 2, 3, {4: "!"}]}])

try:
    x = 5 / 0
except Exception as e:
    shovel_logger.exception("咕", e)



from abc import abstractmethod

import pymysql
from pymysql.cursors import SSDictCursor, SSCursor

from nehushtan.mysql import constant
from nehushtan.mysql.MySQLKit import MySQLKit
from nehushtan.mysql.MySQLQueryResult import MySQLQueryResult


class MySQLSelectionMixin:
    """
    Since 0.3.6
    """

    @abstractmethod
    def get_mysql_kit(self) -> MySQLKit:
        pass

    @abstractmethod
    def generate_sql(self) -> str:
        pass

    def query_for_result_matrix(self, row_type: type):
        result = MySQLQueryResult(self.parse_row_type_to_str(row_type))
        try:
            sql = self.generate_sql()
            result.set_sql(sql)
            if row_type is dict:
                matrix = self.get_mysql_kit().raw_query_for_all_dict_rows(sql)
            else:
                matrix = self.get_mysql_kit().raw_query_for_all_tuple_rows(sql)
            result.set_status(constant.MYSQL_QUERY_STATUS_QUERIED)
            result.append_result_rows(matrix)
        except pymysql.MySQLError as e:
            result.set_status(constant.MYSQL_QUERY_STATUS_ERROR)
            result.set_error(f"MySQL Error {e.__class__} [{e.args[0]}] {e.args[1]}")
        except Exception as pe:
            result.set_status(constant.MYSQL_QUERY_STATUS_ERROR)
            result.set_error(f"Python Error {pe.__class__}: {pe}")
        finally:
            return result

    def query_for_result_as_tuple_of_dict(self):
        return self.query_for_result_matrix(dict)

    def query_for_result_as_tuple_of_tuple(self):
        return self.query_for_result_matrix(tuple)

    def query_for_result_stream(self, row_type: type):
        result = MySQLQueryResult(self.parse_row_type_to_str(row_type))
        try:
            sql = self.generate_sql()
            result.set_sql(sql)
            if row_type is dict:
                cursor = self.get_mysql_kit().get_raw_connection().cursor(SSDictCursor)
            else:
                cursor = self.get_mysql_kit().get_raw_connection().cursor(SSCursor)
            cursor.execute(sql)
            result.set_status(constant.MYSQL_QUERY_STATUS_STREAMING)
            result.set_stream(cursor)
            return result
        except pymysql.MySQLError as e:
            result.set_status(constant.MYSQL_QUERY_STATUS_ERROR)
            result.set_error(f"MySQL Error {e.__class__} [{e.args[0]}] {e.args[1]}")
        except Exception as pe:
            result.set_status(constant.MYSQL_QUERY_STATUS_ERROR)
            result.set_error(f"Python Error {pe.__class__}: {pe}")
        finally:
            return result

    def query_for_result_stream_as_dict(self):
        return self.query_for_result_stream(row_type=dict)

    def query_for_result_stream_as_tuple(self):
        return self.query_for_result_stream(row_type=tuple)

    @staticmethod
    def parse_row_type_to_str(row_type: type):
        if row_type is dict:
            return constant.MYSQL_QUERY_ROW_TYPE_DICT
        elif row_type is tuple:
            return constant.MYSQL_QUERY_ROW_TYPE_TUPLE
        else:
            return constant.MYSQL_QUERY_ROW_TYPE_UNKNOWN



import os
import socket
import time
from threading import Thread, Event

from nehushtan.socket.NehushtanTCPSocketServer import NehushtanTCPSocketServer


class SampleEvent(Event):
    pass


event = SampleEvent()


class QAServer(NehushtanTCPSocketServer):
    def handle_incoming_connection(self, connection: socket.socket, address):
        # peer = connection.getpeername()

        event.set()

        try:
            while True:
                # print(f'[DEBUG] recv-ing from {peer}')
                buffer = connection.recv(1024)
                # print(f'[DEBUG] recv-ed from {peer}: {buffer}')

                if not buffer:
                    # print(f'[FATAL] recv empty so connection might be closed')
                    break

                parts = buffer.split(b'+')
                s = 0
                for part in parts:
                    if part:
                        s += int(part.decode())
                connection.send(f'{s}'.encode())
        except ConnectionResetError as e:
            # print(f'[WARN] ConnectionResetError by peer [{peer}]: {e}')
            pass

        # print(f'[DEBUG] finished with {peer}')


def with_monitor_fork():
    main_pid = os.getpid()
    child_pid = os.fork()
    if child_pid == 0:
        # I am the child
        for i in range(60):
            print(f'[MONITOR] <{i}> total threads: {server.get_thread_manager().check_alive_thread_count()}')
            time.sleep(1)

        print('[MONITOR] OVER')
    else:
        # Main
        server.listen()


def with_monitor_thread(server: NehushtanTCPSocketServer):
    thread = Thread(target=monitor_thread_worker, args=(server,))
    thread.start()


def monitor_thread_worker(server: NehushtanTCPSocketServer):
    while event.wait():
        print(f'[MONITOR] total threads: {server.get_thread_manager().check_alive_thread_count()}')
        event.clear()

    print('[MONITOR] OVER')


if __name__ == '__main__':
    server = QAServer(4354)

    with_monitor_thread(server)

    server.listen()



import json
import math
import os
import sys
import threading
import time
import traceback
from datetime import datetime

import psutil

from nehushtan.logger.NehushtanLogging import NehushtanLogging


class NehushtanFileLogger:
    """
    @since 0.1.25
    Another solution for logging, just use raw writing method to target file.
    """

    def __init__(
            self,
            title='default',
            log_dir: str = None,
            log_level=NehushtanLogging.DEBUG,
            categorize: bool = True,
            date_rotate: bool = True,
            print_higher_than_this_level=NehushtanLogging.CRITICAL,
            record_millisecond=False,
    ):
        self.categorize = categorize
        self.log_dir = log_dir
        self.title = title

        # This logic is since 0.4.19
        if self.categorize:
            x = title.split('/')
            last_title = None
            y = []
            for xx in x:
                if xx:
                    if last_title:
                        y.append(last_title)
                    last_title = xx
            if y and self.log_dir:
                self.log_dir = self.log_dir + '/' + ('/'.join(y))
            if last_title:
                self.title = last_title

        self.log_level = log_level

        self.date_rotate = date_rotate
        self.record_millisecond = record_millisecond
        # self.print_as_well = print_as_well <-- should use print_higher_than_this_level=NehushtanLogging.NOTSET
        # If all, use NOTSET; if none, use FATAL
        if print_higher_than_this_level is True:
            self.print_higher_than_this_level = NehushtanLogging.NOTSET
        elif print_higher_than_this_level is False:
            self.print_higher_than_this_level = NehushtanLogging.CRITICAL
        else:
            self.print_higher_than_this_level = print_higher_than_this_level

    def get_target_file(self):
        if self.log_dir is None:
            return ''

        category_dir = self.log_dir

        # TODO
        # a -> a/a-DATE.log
        # a/b -> a/b-DATE.log
        # a/b/c -> a/b/c-DATE.log

        if self.categorize:
            category_dir = os.path.join(self.log_dir, self.title)

        if not os.path.exists(category_dir):
            os.makedirs(category_dir)

        today = ''
        if self.date_rotate:
            today = time.strftime("%Y%m%d", time.localtime())
            today = f'-{today}'

        target_file = os.path.join(category_dir, f'{self.title}{today}.log')

        return target_file

    def write_raw_line_to_log(self, text: str, level: int = NehushtanLogging.INFO, end=os.linesep):
        """
        Parameter `level` is only used to determine stdout or stderr when file empty.
        Since 0.2.8, Parameter `end` added.
        """
        target_file = self.get_target_file()

        if target_file != '':
            file = open(target_file, 'a')
            file.write(text + end)
            file.flush()
            file.close()

        if target_file == '' or level > self.print_higher_than_this_level:
            if level >= NehushtanLogging.WARNING:
                print(text, file=sys.stderr, end=end)
            else:
                print(text, end=end)

        return self

    @staticmethod
    def get_level_label(level: int):
        return NehushtanLogging.get_label_of_level(level)

    def write_formatted_line_to_log(self, level: int, message: str, extra=None, hide_extra: bool = False):
        """
        Since 0.3.7 add `hide_extra`
        """
        if level < self.log_level:
            return self
        time_format_string = "%Y-%m-%d %H:%M:%S"
        if self.record_millisecond:
            time_format_string += '.%f'
        # now = time.strftime(time_format_string, time.localtime())
        now = datetime.now().strftime(time_format_string)
        level_label = NehushtanFileLogger.get_level_label(level)
        pid = os.getpid()
        thread = threading.currentThread()
        line = f'{now} <{self.title}> [{level_label}] <{pid}:{thread.getName()}> {message}'
        if not hide_extra:
            extra_json = self.ensure_extra_as_dict(extra)
            line += f' | {extra_json}'
        return self.write_raw_line_to_log(line, level)

    def debug(self, message: str, extra=None):
        return self.write_formatted_line_to_log(NehushtanLogging.DEBUG, message, extra)

    def info(self, message: str, extra=None):
        return self.write_formatted_line_to_log(NehushtanLogging.INFO, message, extra)

    def notice(self, message: str, extra=None):
        return self.write_formatted_line_to_log(NehushtanLogging.NOTICE, message, extra)

    def warning(self, message: str, extra=None):
        return self.write_formatted_line_to_log(NehushtanLogging.WARNING, message, extra)

    def error(self, message: str, extra=None):
        return self.write_formatted_line_to_log(NehushtanLogging.ERROR, message, extra)

    def exception(self, message: str, exception: BaseException):
        just_the_string = self.get_traceback_info_from_exception(exception)
        return self.write_formatted_line_to_log(NehushtanLogging.ERROR, message, f'{type(exception).__name__}') \
            .write_raw_line_to_log(just_the_string, NehushtanLogging.ERROR)

    def critical(self, message: str, extra=None):
        return self.write_formatted_line_to_log(NehushtanLogging.CRITICAL, message, extra)

    def log_progress(
            self,
            title: str,
            done_task_count: int,
            total_task_count: int = 100,
            progress_bar_length: int = 20,
            desc: str = None,
            level: int = NehushtanLogging.NOTICE
    ):
        """
        Since 0.3.7
        """
        percent = 100.0 * done_task_count / total_task_count
        bar = ''
        done_bar_chars = math.floor(progress_bar_length * percent / 100.0)
        for i in range(done_bar_chars):
            bar += '='
        for j in range(progress_bar_length - done_bar_chars):
            bar += '-'

        content = f'{title}: {percent:2.2f}% ({done_task_count}/{total_task_count}) [{bar}]'
        if desc is not None:
            content += f' {desc}'
        return self.write_formatted_line_to_log(
            level,
            content,
            hide_extra=True
        )

    def log_current_memory_usage_of_process(self, pid: int = None, level: int = NehushtanLogging.INFO):
        """
        Since 0.2.10
        Filed `pid` would use os.getpid() for None.
        """
        memory_usage = psutil.Process(pid=pid).memory_info()
        self.write_formatted_line_to_log(
            level,
            'Current Memory Usage Snapshot (in MB)',
            {'rss': memory_usage.rss / 1024.0 / 1024.0, 'vms': memory_usage.vms / 1024.0 / 1024.0}
        )

    def log_current_memory_usage_of_object(self, target_name: str, target, level: int = NehushtanLogging.INFO):
        """
        Since 0.2.10
        """
        memory_usage = sys.getsizeof(target)
        self.write_formatted_line_to_log(
            level,
            f'Current Memory Usage used by {target_name} (in MB)',
            memory_usage / 1024.0 / 1024.0
        )

    @staticmethod
    def ensure_extra_as_dict(extra):
        """
        Since 0.1.25, add ensure_ascii as False to allow unicode chars
        """
        return json.dumps(extra, default=lambda inner_x: inner_x.__str__(), ensure_ascii=False)

    def get_args_json_to_clone(self):
        """
        Since 0.2.15
        """
        return json.dumps([
            self.title,
            self.log_dir,
            self.log_level,
            self.categorize,
            self.date_rotate,
            self.print_higher_than_this_level,
            self.record_millisecond,
        ])

    @staticmethod
    def build_instance_from_args_json(args_json: str):
        """
        Since 0.2.15
        """
        args = json.loads(args_json)
        if type(args) is not list and type(args) is not tuple:
            raise ValueError('String `args_json` should be an array in JSON format')
        return NehushtanFileLogger(*args)

    @staticmethod
    def get_traceback_info_from_exception(exception: BaseException) -> str:
        """
        Since 0.4.4
        """
        return ''.join(
            traceback.format_exception(
                etype=type(exception),
                value=exception,
                tb=exception.__traceback__
            )
        )



from nehushtan.queue.NehushtanQueue import NehushtanQueue
from nehushtan.queue.NehushtanQueueDelegate import NehushtanQueueDelegate
from nehushtan.queue.NehushtanQueueTaskDelegate import NehushtanQueueTaskDelegate


class Test2NehushtanQueue(NehushtanQueue):
    def __init__(self, delegate: NehushtanQueueDelegate, task_delegate: NehushtanQueueTaskDelegate):
        super().__init__(delegate, task_delegate)



from nehushtan.mysql.MySQLAnyTable import MySQLAnyTable
from nehushtan.mysql.MySQLKit import MySQLKit
from nehushtan.mysql.MySQLKitConfig import MySQLKitConfig
from tests.config import MYSQL_CONFIG

table = MySQLAnyTable(
    MySQLKit(MySQLKitConfig(MYSQL_CONFIG)),
    'a'
)

fields = ['id', 'data']
matrix = [[1, 'e'], (3, '3'), [4, '5']]

result = table.replace_many_rows_with_matrix(fields, matrix)
print(result.get_status(), result.get_error(), result.get_sql())

dicts = [{'id': 2, 'data': '55'}, {'id': 4, 'data': 3.3}]

result = table.insert_many_rows_with_dicts(dicts, on_duplicate_key_update_rows={'data': 'data+1'})
print(result.get_status(), result.get_error(), result.get_sql())

# result = table.update_rows(conditions=[MySQLCondition.make_equal('id', 3), 3], modifications={"data": 4})
# print(result.get_status(), result.get_error(), result.get_sql())
#
# result = table.update_rows(conditions=MySQLCondition.make_equal('id', 3), modifications={"data": 4})
# print(result.get_status(), result.get_error(), result.get_sql())



import importlib
import uuid
import warnings


class CommonHelper:
    def __init__(self):
        pass

    @staticmethod
    def read_target(target, keychain: tuple, default: any = None):
        if type(target) is dict:
            return CommonHelper.read_dictionary(target, keychain, default)
        elif type(target) is list or type(target) is tuple:
            return CommonHelper.read_array(target, keychain, default)
        else:
            return default

    @staticmethod
    def read_dictionary(dictionary: dict, keychain: tuple, default: any = None):
        if len(keychain) <= 0:
            return default
        elif len(keychain) == 1:
            return dictionary.get(keychain[0], default)
        else:
            current_key = keychain[0]
            if dictionary.keys().__contains__(current_key):
                sub_dictionary = dictionary.get(current_key, None)
                if type(sub_dictionary) is dict:
                    return CommonHelper.read_dictionary(sub_dictionary, keychain[1:], default)
                elif type(sub_dictionary) is tuple or type(sub_dictionary) is list:
                    return CommonHelper.read_array(sub_dictionary, keychain[1:], default)
                else:
                    return default
            else:
                return default

    @staticmethod
    def read_array(array: tuple, keychain: tuple, default: any = None):
        if len(keychain) <= 0:
            return default
        elif len(keychain) == 1:
            current_key = int(keychain[0])
            if len(array) > current_key:
                return array[int(keychain[0])]
            else:
                return default
        else:
            current_key = int(keychain[0])
            if len(array) > current_key:
                sub_array = array[current_key]
                if type(sub_array) is tuple or type(sub_array) is list:
                    return CommonHelper.read_array(sub_array, keychain[1:], default)
                elif type(sub_array) is dict:
                    return CommonHelper.read_dictionary(sub_array, keychain[1:], default)
                else:
                    return default
            else:
                return default

    @staticmethod
    def write_dictionary(target_dict: dict, keychain: tuple, value: any):
        keychain_length = len(keychain)
        if keychain_length > 1:
            current_key = keychain[0]
            current_target = target_dict.get(current_key)
            if type(current_target) is not dict:
                # Since 0.2.14
                # a great change: all non-dict-type entry would be cleared as an empty dict
                target_dict[current_key] = {}
            CommonHelper.write_dictionary(target_dict[current_key], keychain[1:], value)
        elif keychain_length == 1:
            target_dict[keychain[0]] = value

        return target_dict

    @staticmethod
    def class_with_class_path(module_path: str, class_name: str = None):
        """
        Since 0.1.22
        Since 0.2.19 When the class name is the same with PY file name, `class_name` is optional.

        For a/b.py -> class b
        class_with_namespace is like 'package.sub_package.class', 'a.b'
        class_name is 'b'
        return a CLASS definition, to be used with parameters to make instance
        """
        module = importlib.import_module(module_path)
        if class_name is None:
            class_name = module_path.split('.')[-1]
        a_class = getattr(module, class_name)
        return a_class

    @staticmethod
    def class_with_module_and_name(module_base: str, sub_module_name: str):
        """
        Since 0.1.21
        Since 0.2.19 It is not so convinence to use, consider using `class_with_class_path`.

        For a/b.py -> class b
        module_base is a
        sub_module_name is b
        """
        warnings.warn('Use `class_with_class_path` instead.')

        module = __import__(module_base)
        a_class = getattr(module, sub_module_name)
        return a_class

    @staticmethod
    def generate_random_uuid_hex():
        """
        Generate a random UUID.
        Since 0.4.15
        """
        return uuid.uuid4().hex



import asyncio

import websockets

from nehushtan.logger.NehushtanFileLogger import NehushtanFileLogger
from tests.ws.TestWebsocketAgent import TestWebsocketAgent
from tests.ws.TestWebsocketRegisterAgent import TestWebsocketRegisterAgent


def daemon():
    request_handle_logger = NehushtanFileLogger('ws-request', '/Users/leqee/code/nehushtan/log/ws')
    broadcast_logger = NehushtanFileLogger('ws-broadcast', '/Users/leqee/code/nehushtan/log/ws')

    TestWebsocketAgent.register_agent = TestWebsocketRegisterAgent()
    agent = TestWebsocketAgent(
        broadcast_sleep_seconds=5,
        request_handle_logger=request_handle_logger,
        broadcast_logger=broadcast_logger
    )
    start_server = websockets.serve(agent.ws_request_handler, "localhost", 8080)
    asyncio.get_event_loop().create_task(agent.ws_broadcast_handler())
    asyncio.get_event_loop().run_until_complete(start_server)
    asyncio.get_event_loop().run_forever()


if __name__ == '__main__':
    daemon()



import socket
import time

from nehushtan.socket.NehushtanTCPSocketServer import NehushtanTCPSocketServer


class EchoServer(NehushtanTCPSocketServer):
    def should_terminate(self) -> bool:
        return False

    def handle_incoming_connection(self, connection: socket.socket, address):
        print(f'[DEBUG] handle_incoming_connection start for {address}')

        peer = connection.getpeername()
        try:
            while True:
                received = connection.recv(1024)
                if not received:
                    print(f'[FATAL] connection received empty from {peer} may be closed')
                    break

                print(f'{connection.getpeername()}: {received}')

                connection.send(received)
                time.sleep(2)
        except ConnectionResetError as e:
            print(f'[ERROR] ConnectionResetError with {peer}: {e}')


if __name__ == '__main__':
    EchoServer(4444).listen()



from nehushtan.httpd.exceptions.NehushtanHTTPError import NehushtanHTTPError


class NehushtanRequestParameterError(NehushtanHTTPError):
    """
    When a request cannot be processed due to errors in the parameters
    """

    TYPE_QUERY = "QUERY"
    TYPE_BODY = "BODY"
    TYPE_HEADER = "HEADER"
    TYPE_COOKIE = "COOKIE"

    DESC_NOT_SET = 'NOT_SET'
    DESC_FORMAT_ERROR = "FORMAT_ERROR"

    def __init__(self, key: str, key_type: str, desc: str = 'Undetermined Error', http_code: int = 400):
        super().__init__(
            f'Request Key [{key}] of Type [{key_type}]: {desc}',
            http_code
        )
        self.key = key
        self.key_type = key_type
        self.desc = desc



import time
from abc import abstractmethod
from typing import List

from nehushtan.helper.CommonHelper import CommonHelper
from nehushtan.logger.NehushtanFileLogger import NehushtanFileLogger
from nehushtan.queue.NehushtanQueueTask import NehushtanQueueTask


class NehushtanQueueDelegate:
    QUEUE_RUNTIME_COMMAND_PAUSE = "PAUSE"
    QUEUE_RUNTIME_COMMAND_CONTINUE = "CONTINUE"
    QUEUE_RUNTIME_COMMAND_STOP = "STOP"
    # QUEUE_RUNTIME_COMMAND_FORCE_STOP = "FORCE-STOP"
    # QUEUE_RUNTIME_COMMAND_RESTART = "RESTART"
    # QUEUE_RUNTIME_COMMAND_FORCE_RESTART = "FORCE-RESTART"

    CONFIG_KEY_POOL_CAPACITY = 'POOL_CAPACITY'
    CONFIG_KEY_SLEEP_TIME_FOR_FREE = 'SLEEP_TIME_FOR_FREE'
    CONFIG_KEY_SLEEP_TIME_FOR_BUSY = 'SLEEP_TIME_FOR_BUSY'
    CONFIG_KEY_SLEEP_TIME_FOR_PAUSE = 'SLEEP_TIME_FOR_PAUSE'

    def __init__(self, config_dictionary: dict = None, logger: NehushtanFileLogger = None):
        """
        @since 0.1.25 logger changed to NehushtanFileLogger
        """
        if config_dictionary is None:
            config_dictionary = {}
        self.config_dictionary = config_dictionary

        if logger is None:
            logger = NehushtanFileLogger(f"{self.__class__.__name__}-Logger", )
        self.logger = logger

        self.latest_command = NehushtanQueueDelegate.QUEUE_RUNTIME_COMMAND_CONTINUE

    def read_config_of_delegate(self, keychain: tuple, default):
        return CommonHelper.read_dictionary(self.config_dictionary, keychain, default)

    def get_configured_pool_capacity(self) -> int:
        return self.read_config_of_delegate((NehushtanQueueDelegate.CONFIG_KEY_POOL_CAPACITY,), 1)

    def sleep_for_configured_time(self, config_key: str):
        """
        config_key might be amongst `CONFIG_KEY_SLEEP_TIME_FOR_*`
        """
        x = self.read_config_of_delegate((config_key,), 1)
        time.sleep(x)

    def read_latest_command(self) -> str:
        """
        Override this to fetch command from real pipe.
        This works in MASTER
        """
        self.latest_command = self.QUEUE_RUNTIME_COMMAND_CONTINUE
        return self.latest_command

    def when_loop_starts(self):
        """
        To be executed before loop cycle starts
        This works in MASTER
        """
        pass

    @abstractmethod
    def when_loop_reports_error(self, error_message: str, cause_exception: Exception = None):
        """
        This works in MASTER
        Since 0.4.4 Added cause_exception, could be expressed by NehushtanFileLogger
        """
        pass

    def is_runnable(self):
        """
        If not runnable, the daemon loop would sleep.
        This works in MASTER
        """
        return self.latest_command == self.QUEUE_RUNTIME_COMMAND_CONTINUE

    def should_terminate(self):
        """
        Tell daemon loop to exit.
        This works in MASTER
        """
        return (
            self.QUEUE_RUNTIME_COMMAND_STOP,
            # self.QUEUE_RUNTIME_COMMAND_FORCE_STOP,
            # self.QUEUE_RUNTIME_COMMAND_RESTART,
            # self.QUEUE_RUNTIME_COMMAND_FORCE_RESTART,
        ).__contains__(self.latest_command)

    @abstractmethod
    def when_loop_terminates(self):
        """
        When the loop gets ready to terminate by shouldTerminate instructed, execute this
        This works in MASTER
        """
        pass

    def when_loop_should_not_run(self):
        """
        Sleep for a certain while.
        This works in MASTER
        """
        self.sleep_for_configured_time(self.CONFIG_KEY_SLEEP_TIME_FOR_PAUSE)

    def before_seeking_next_tasks(self):
        """
        This works in MASTER
        It is a hook before `check_next_task`.
        A scnerio is, here delegate prepares a list of task candidates,
        And be consumed in `check_next_task` in order.
        Since 0.2.19
        """
        pass

    @abstractmethod
    def check_next_task_candidates(self) -> List[NehushtanQueueTask]:
        """
        This works in MASTER
        if none newly found... raise NoNextTaskSituation!
        Since 0.3.0
        """
        pass

    def when_no_task_to_do(self):
        """
        When the loop cannot check for a task to do next, execute this
        This works in MASTER
        """
        self.sleep_for_configured_time(self.CONFIG_KEY_SLEEP_TIME_FOR_FREE)

    @abstractmethod
    def when_task_not_executable(self, task: NehushtanQueueTask):
        """
        this is done before fork in pooled style
        This works in MASTER
        """
        pass

    def when_pool_is_full(self):
        """
        This works in MASTER
        """
        self.logger.warning('Pool is full, sleep for a while', {'size': self.get_configured_pool_capacity()})
        self.sleep_for_configured_time(self.CONFIG_KEY_SLEEP_TIME_FOR_BUSY)

    def when_worker_process_created(self, task_reference, note: str = ''):
        """
        This works in MASTER
        """
        pass

    def when_worker_process_confirmed_dead(self, task_reference, detail=None):
        """
        This works in MASTER
        """
        pass

    # def should_wait_when_pool_is_full(self):
    #     """
    #     This works in MASTER
    #     """
    #     return True

    def should_wait_for_all_workers_before_terminating(self):
        """
        This works in MASTER
        """
        return True

    @abstractmethod
    def should_kill_any_worker_processes(self) -> list:
        """
        This works in MASTER
        return a list of Task Reference
        """
        pass

    @abstractmethod
    def when_killed_worker_process(self, task_reference, not_found=False, worker_pid: int = None):
        """
        This works in MASTER
        """
        pass

    @abstractmethod
    def handle_command_queue(self) -> int:
        """
        This works in MASTER
        This may update the data for `should_kill_any_worker_processes` for loop,
        or do other jobs
        Since 0.2.18 should return an integer, as the count of commands handled; so, zero for no command to do.
        """
        pass

    def register_queue_news(self, title: str, content: str, running_worker_count: int, busy_rate: float):
        """
        Since 0.2.9
        This works in MASTER
        It provides a method to let delegate know what daemon happened, other than the `when_*` methods.
        By default, the news would be ignored.
        It could be used to write a message queue, print to heartbeat logs, etc.
        """
        pass
